# ---------------------------------------------------------
# Copyright (c) vuepy.org. All rights reserved.
# ---------------------------------------------------------
from __future__ import annotations

import ast
import dataclasses
import enum
from _ast import Load
from _ast import Name
from _ast import Starred
from _ast import keyword
from dataclasses import field
from typing import Dict
from typing import List
from typing import Tuple

from vuepy.compiler_core.component_expr import VueCompExpr
from vuepy.runtime.core.api_setup_helpers import defineModel


@dataclasses.dataclass
class VForAst:
    iter: str
    target: str
    idx: str = None

    @classmethod
    def parse(cls, exp):
        """
        (i, target) in iter
        :param exp:
        :return:
        """
        if not exp:
            return None

        try:
            i_target, iters = [i.strip() for i in exp.split(' in ')]
            i_target = [i.strip() for i in i_target.strip('()').split(',')]
            if len(i_target) == 1:
                i, target = None, i_target[0]
            else:
                i, target = i_target[0], i_target[1]
        except Exception as err:
            msg = f"parse v-for='{exp}' failed, {err}"
            raise ValueError(msg)

        return cls(iter=iters, target=target.strip(), idx=i)


@dataclasses.dataclass
class VForScopes:
    idxs: Tuple[int]
    vars: dict

    def to_ns(self):
        return self.vars


class NodeAstType(enum.Enum):
    HTML = 'html'
    WIDGET = 'widget'


@dataclasses.dataclass
class NodeAst:
    tag: str
    attrs: dict = field(default_factory=dict)
    parent: "NodeAst" = None
    children: List["NodeAst"] = field(default_factory=list)
    plain: bool = False
    v_for: VForAst = None
    v_for_scopes: VForScopes = None
    for_processed: bool = False
    type: NodeAstType = NodeAstType.WIDGET

    def add_child(self, child):
        self.children.append(child)


@dataclasses.dataclass
class VForNodeAst(NodeAst):

    @property
    def children_flat(self) -> List["NodeAst"]:
        ret = []

        def _travel(children):
            if not isinstance(children, list):
                ret.append(children)
                return

            for child in children:
                _travel(child)

        _travel(self.children)
        return ret


class VueCompAst:
    V_IF = 'v-if'
    V_ELSE_IF = 'v-else-if'
    V_ELSE = 'v-else'
    V_FOR = 'v-for'
    V_SHOW = 'v-show'
    V_SLOT = 'v-slot:'
    V_SLOT_ABBR = '#'
    V_JS_LINK = 'v-js-link'
    V_REF = 'ref'
    V_MODEL = 'v-model:'
    V_MODEL_DEFAULT = 'v-model'
    V_HTML = 'v-html'
    V_ON = 'v-on:'
    V_ON_ABBR = '@'
    V_BIND = 'v-bind:'
    V_BIND_ABBR = ':'

    LAYOUT_ATTRS = {'width', 'height', 'padding', 'border'}

    # todo 和h函数的参数保持一致
    def __init__(self, tag):
        self.tag = tag
        # todo rename to attrs
        self.kwargs = {}
        self.layout = {}

        self.v_if: VueCompExpr = None
        self.v_else_if: VueCompExpr = None
        self.v_else = False
        self.v_show: VueCompExpr = None
        self.v_binds: Dict[str, VueCompExpr] = {}
        self.v_model: List[(str, str)] = []
        self.v_html = None
        self.v_on: Dict[str, VueCompExpr] = {}
        self.v_slot = None
        self.v_ref = None

    @classmethod
    def is_v_if(cls, attr):
        return attr == cls.V_IF

    @classmethod
    def is_v_else_if(cls, attr):
        return attr == cls.V_ELSE_IF

    @classmethod
    def is_v_else(cls, attr):
        return attr == cls.V_ELSE

    @classmethod
    def is_v_show(cls, attr):
        return attr == cls.V_SHOW

    @classmethod
    def is_v_slot(cls, attr):
        return attr.startswith(cls.V_SLOT)

    @classmethod
    def is_v_slot_addr(cls, attr):
        return attr.startswith(cls.V_SLOT_ABBR)

    @classmethod
    def is_v_ref(cls, attr):
        return attr == cls.V_REF

    @classmethod
    def is_v_model_default(cls, attr):
        return attr == cls.V_MODEL_DEFAULT

    @classmethod
    def is_v_model(cls, attr):
        return attr.startswith(cls.V_MODEL)

    @classmethod
    def is_v_html(cls, attr):
        return attr == cls.V_HTML

    @classmethod
    def is_v_on(cls, attr):
        return attr.startswith(cls.V_ON)

    @classmethod
    def is_v_on_abbr(cls, attr):
        return attr.startswith(cls.V_ON_ABBR)

    @classmethod
    def is_v_bind(cls, attr):
        return attr.startswith(cls.V_BIND)

    @classmethod
    def is_v_bind_abbr(cls, attr):
        return attr.startswith(cls.V_BIND_ABBR)

    @classmethod
    def is_layout(cls, attr):
        return attr in cls.LAYOUT_ATTRS

    @classmethod
    def transform(cls, tag, attr_dict):
        comp = cls(tag)
        for attr, value in attr_dict.items():
            if cls.is_v_if(attr):
                comp.v_if = VueCompExpr.parse(value)

            elif cls.is_v_else_if(attr):
                comp.v_else_if = VueCompExpr.parse(value)

            elif cls.is_v_else(attr):
                comp.v_else = True

            elif cls.is_v_show(attr):
                comp.v_show = VueCompExpr.parse(value)

            elif cls.is_v_bind(attr):
                attr = attr.split(cls.V_BIND, 1)[1]
                comp.v_binds[attr] = VueCompExpr.parse(value)

            elif cls.is_v_bind_abbr(attr):
                attr = attr.split(cls.V_BIND_ABBR, 1)[1]
                comp.v_binds[attr] = VueCompExpr.parse(value)

            elif cls.is_v_model_default(attr):
                comp.v_model.append((defineModel.DEFAULT_KEY, value))

            elif cls.is_v_model(attr):
                _define_model_key = attr.split(cls.V_MODEL, 1)[1]
                comp.v_model.append((_define_model_key, value))

            elif cls.is_v_html(attr):
                comp.v_html = value

            elif cls.is_v_on(attr) or cls.is_v_on_abbr(attr):
                _prefix = cls.V_ON if cls.is_v_on(attr) else cls.V_ON_ABBR
                event = attr.split(_prefix, 1)[1]
                func_ast = VueCompExpr.parse(value)
                if isinstance(func_ast.exp_ast.body, (ast.Name, ast.Attribute)):
                    exp_ast = ast.Expression(
                        ast.Call(
                            func=func_ast.exp_ast.body,
                            args=[Starred(value=Name(id='__vp_args', ctx=Load()), ctx=Load())],
                            keywords=[keyword(value=Name(id='__vp_kwargs', ctx=Load()))],
                        )
                    )
                    ast.fix_missing_locations(exp_ast)
                    func_ast.exp_ast = exp_ast

                comp.v_on[event] = func_ast

            elif tag == 'template' and cls.is_v_slot(attr):
                comp.v_slot = attr.split(cls.V_SLOT, 1)[1]

            elif tag == 'template' and cls.is_v_slot_addr(attr):
                comp.v_slot = attr.split(cls.V_SLOT_ABBR, 1)[1]

            elif cls.is_v_ref(attr):
                comp.v_ref = value

            elif cls.is_layout(attr):
                comp.layout[attr] = value

            else:
                comp.kwargs[attr] = value

        if comp.layout:
            comp.kwargs['layout'] = comp.layout

        return comp



# coding: utf-8
# ---------------------------------------------------------
# Copyright (c) vuepy.org. All rights reserved.
# ---------------------------------------------------------
from vuepy.compiler_sfc.codegen import VueComponent
from vuepy.reactivity.computed import computed
from vuepy.reactivity.effect_scope import effectScope
from vuepy.reactivity.effect_scope import getCurrentScope
from vuepy.reactivity.effect_scope import onScopeDispose
from vuepy.reactivity.reactive import isProxy
from vuepy.reactivity.reactive import isReactive
from vuepy.reactivity.reactive import reactive
from vuepy.reactivity.reactive import toRaw
from vuepy.reactivity.ref import isRef
from vuepy.reactivity.ref import ref
from vuepy.reactivity.ref import shallowRef
from vuepy.reactivity.ref import toRef
from vuepy.reactivity.ref import toRefs
from vuepy.reactivity.ref import toValue
from vuepy.reactivity.ref import triggerRef
from vuepy.reactivity.ref import unref
from vuepy.reactivity.watch import watch
from vuepy.reactivity.watch import watchEffect
from vuepy.runtime.core.api_create_app import App
from vuepy.runtime.core.api_create_app import VueOptions
from vuepy.runtime.core.api_create_app import VuePlugin
from vuepy.runtime.core.api_create_app import create_app
from vuepy.runtime.core.api_lifecycle import onBeforeMount
from vuepy.runtime.core.api_lifecycle import onMounted
from vuepy.runtime.core.api_setup_helpers import defineEmits
from vuepy.runtime.core.api_setup_helpers import defineModel
from vuepy.runtime.core.api_setup_helpers import defineProps
from vuepy.runtime.core.import_sfc import import_sfc
from vuepy.utils.appstore import VuepyAppStore
from vuepy.version import VERSION

__version__ = VERSION

__all__ = [
    'App',
    'VuepyAppStore',
    'create_app',
    'import_sfc',
    'VueComponent',
    'VuePlugin',
    'VueOptions',
    'ref',
    'computed',
    'reactive',
    'watch',
    'watchEffect',
    'shallowRef',
    'triggerRef',
    'toRaw',
    'effectScope',
    'getCurrentScope',
    'onScopeDispose',
    'isRef',
    'unref',
    'toRef',
    'toValue',
    'toRefs',
    'isProxy',
    'isReactive',
    'defineEmits',
    'defineModel',
    'defineProps',
    'onBeforeMount',
    'onMounted',
]



# ---------------------------------------------------------
# Copyright (c) vuepy.org. All rights reserved.
# ---------------------------------------------------------
from __future__ import annotations

import abc
import dataclasses
from dataclasses import field
from typing import Any
from typing import Type
from typing import Union

import ipywidgets as widgets
from IPython.display import clear_output
from IPython.display import display

from vuepy import log
from vuepy.compiler_core.options import CompilerOptions
from vuepy.compiler_sfc.codegen import Dom
from vuepy.compiler_sfc.codegen import SFC
from vuepy.compiler_sfc.codegen import SFCFactory
from vuepy.compiler_sfc.codegen import VueComponent
from vuepy.utils.common import Record

logger = log.getLogger()


@dataclasses.dataclass
class AppConfig:
    """应用的配置设定"""
    # error_handler: ErrorHandler
    # warn_handler: WarningHandler
    # option_merge_strategies: dict[str, OptionMergeFunction] = {}
    globalProperties: Record = field(default_factory=Record)
    compilerOptions: CompilerOptions = field(default_factory=CompilerOptions)
    performance: bool = False


class Directive:
    pass


@dataclasses.dataclass
class AppContext:
    app: 'App'
    config: AppConfig
    directives: dict[str, Directive]
    provides: dict[str, Any]
    components: dict[str, VueComponent] = field(default_factory=dict)


class App:
    components = {}

    def __init__(self, root_component: RootComponent, debug=False):
        self.config: AppConfig = AppConfig()

        self._installed_plugins = []
        self._props: dict = {}
        self._context: AppContext = AppContext(self, self.config, {}, {})
        self._instance: VueComponent = None

        if isinstance(root_component, dict):
            root_component = VueOptions(**root_component)

        props = {}
        context = {}
        if isinstance(root_component, SFCFactory):
            self.root_component: SFC = root_component.gen(props, context, self)
        else:
            raise ValueError(
                f"root_component only support {RootComponent}, {type(root_component)} found."
            )
        self.debug = debug

        self._components = {}

        self.document: Document = Document()
        self.dom = None

        self._proxy_methods()

    @property
    def version(self):
        import vuepy
        return vuepy.__version__

    def _call_if_callable(self, func):
        if callable(func):
            func(self)

    # def __getattr__(self, key):
    #     if key == '_data':
    #         return super().__getattribute__(key)
    #
    #     if key in self._data:
    #         return self._data[key]
    #
    #     return super().__getattribute__(key)
    #
    # def __setattr__(self, key, value):
    #     if key == '_data':
    #         return super().__setattr__(key, value)
    #
    #     if key in self._data:
    #         self._data[key] = value
    #         return
    #     return super().__setattr__(key, value)

    def _proxy_methods(self):
        pass

    def render(self):
        logger.info('App render start.')
        if not isinstance(self.root_component, SFC):
            raise ValueError(f"render failed, root_component type {type(self.root_component)}.")

        # self.dom = self.options.render({}, self._props, self._data)
        self.dom = self.root_component.render()
        # with self.options.el:
        # with self._container:
        #     clear_output(True)
        #     display(self.dom)
        logger.info('App render end.')
        return self.dom

    def component(self, name: str, comp: 'VueComponent' = None) -> App:
        """
        query component(name) -> Component | None
        register component(name, comp) -> self

        :param name:
        :param comp:
        :return: self
        """
        # query
        if comp is None:
            return self._components.get(name, self.components.get(name, None))

        # register
        self._components[name] = comp
        return self

    def directive(self, name: str, directive: Directive = None) -> Directive | "App":
        """
        query directive(name) -> Directive | None
        register directive(name) -> self

        :param name:
        :param directive:
        :return:
        """
        pass

    def use(self, plugin: Type["VuePlugin"], options: dict = None) -> "App":
        """
        install plugin.

        :param plugin: 插件本身
        :param options: 要传递给插件的选项
        :return: self
        """
        if plugin in self._installed_plugins:
            return self

        plugin.install(self, options)
        return self

    def mount(self, el=None):
        # self._call_if_callable(self.options.before_mount)
        self.render()
        # self._call_if_callable(self.options.mounted)

        self.document.body.appendChild(self.dom)

        return self.document


class VuePlugin:
    @classmethod
    @abc.abstractmethod
    def install(cls, app: App, options: dict):
        pass


VueOptions = SFCFactory

RootComponent = Type[Union[Type[VueComponent], SFCFactory, dict]]


def create_app(root_component: RootComponent, use_wui=True, **root_props) -> App:
    """创建一个应用实例
    app = create_app(App)
    app = create_app({})

    :param use_wui:
    :param root_component:
    :param root_props:
    :return:
    """
    debug = root_props.get('debug', False)
    app = App(root_component, debug)
    if use_wui:
        from ipywui import wui
        app.use(wui)
    return app


class Document(widgets.VBox):
    """
    https://developer.mozilla.org/en-US/docs/Web/API/Document
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.body = Dom()
        self.children = (self.body,)



# ---------------------------------------------------------
# Copyright (c) vuepy.org. All rights reserved.
# ---------------------------------------------------------
import logging

import ipywidgets as widgets
from IPython.display import display

from vuepy.reactivity import config

LOGGER_NAME = 'vuepy'
logout = widgets.Output()


class OutputWidgetHandler(logging.Handler):
    """ Custom logging handler sending logs to an output widget """

    def __init__(self, out, *args, **kwargs):
        super(OutputWidgetHandler, self).__init__(*args, **kwargs)
        self.out = out

    def emit(self, record):
        """ Overload of logging.Handler method """
        formatted_record = self.format(record)
        new_output = {
            'name': 'stdout',
            'output_type': 'stream',
            'text': formatted_record + '\n'
        }
        self.out.outputs = (new_output,) + self.out.outputs

    def show_logs(self):
        """ Show the logs """
        display(self.out)

    def clear_logs(self):
        """ Clear the current logs """
        self.out.clear_output()


def getLogger(name=LOGGER_NAME):
    return logging.getLogger(name)


def init():
    handler = OutputWidgetHandler(logout)
    formatter = logging.Formatter(
        '%(asctime)s [%(levelname)s] %(filename)s:%(funcName)s:%(lineno)s - %(message)s'
    )
    handler.setFormatter(formatter)

    logger = getLogger(LOGGER_NAME)
    logger.addHandler(handler)
    logger.setLevel(config.LOG_LEVEL)
    logger.propagate = False


init()



# ---------------------------------------------------------
# Copyright (c) vuepy.org. All rights reserved.
# ---------------------------------------------------------
from __future__ import annotations

from vuepy.compiler_sfc.codegen import SFCFactory
from vuepy.compiler_sfc.compile import SFCFile


def import_sfc(sfc_file, raw_content=False):
    """
    import Component from sfc_file or raw_content fo sfc

    :param sfc_file:
    :param raw_content:
    :return:
    """
    sfc_file = SFCFile.loads(sfc_file) if raw_content else SFCFile.load(sfc_file)

    return SFCFactory(**{
        'setup': sfc_file.setup_fn,
        'template': sfc_file.template,
        '_file': sfc_file.file,
    })



# coding: utf-8
# ---------------------------------------------------------
# Copyright (c) vuepy.org. All rights reserved.
# ---------------------------------------------------------
import argparse

from vuepy.cli.create_vue import add_arg_parser


def main():
    parser = argparse.ArgumentParser(prog='PROG')
    add_arg_parser(parser)
    args = parser.parse_args()
    try:
        args.func(args)
    except Exception as e:
        print(f"error {e}")
        parser.print_help()
        exit(1)


if __name__ == '__main__':
    main()



# coding: utf-8
# ---------------------------------------------------------
# Copyright (c) vuepy.org. All rights reserved.
# ---------------------------------------------------------

class FactoryMeta(type):
    def __init__(cls, name, bases, attrs):
        super().__init__(name, bases, attrs)
        cls._registry = {}

    def register(cls, name=None):
        def _(sub_cls):
            _name = name or (sub_cls.name() if callable(sub_cls.name) else sub_cls.name)
            cls._registry[_name] = sub_cls
            return sub_cls
        return _

    def get_all_registry(cls):
        return cls._registry

    def create(cls, *args, **kwargs):
        return cls(*args, **kwargs)



# ---------------------------------------------------------
# Copyright (c) vuepy.org. All rights reserved.
# ---------------------------------------------------------
import enum


class IterateKey(enum.Enum):
    DICT = 'dict'
    LIST = 'list'
    SET = 'set'

    @classmethod
    def get_key(cls, target):
        if isinstance(target, dict):
            return IterateKey.DICT
        elif isinstance(target, list):
            return IterateKey.LIST
        elif isinstance(target, set):
            return IterateKey.SET
        else:
            return None


class ErrorCodes(enum.Enum):
    WATCH_GETTER = 2
    WATCH_CALLBACK = 3
    WATCH_CLEANUP = 4



# ---------------------------------------------------------
# Copyright (c) vuepy.org. All rights reserved.
# ---------------------------------------------------------
from typing import List
from typing import Tuple

from ipywui.widgets import WidgetCssStyle
from ipywui.widgets.custom.message import MessageService
from vuepy import App
from vuepy import VueComponent
from vuepy.runtime.core.api_create_app import VuePlugin
from vuepy.utils.factory import FactoryMeta


class wui(VuePlugin, metaclass=FactoryMeta):
    @classmethod
    def install(cls, app: App, options: dict):
        components = cls.get_all_registry()
        for name, component in components.items():
            app.component(name, component)

        app.message = MessageService(app_instance=app)
        app.document.body.appendLeftChild(app.message.widget)


class IPywidgetsComponent(VueComponent):
    STYLE_ATTR = 'style'
    PARAMS_STORE_TRUE: List[Tuple[str, bool]] = []

    def update_style(self, kw):
        styles = kw.pop(self.STYLE_ATTR, None)
        if styles:
            kw.update(WidgetCssStyle.convert_css_style_to_widget_style_and_layout(styles))

    def _process_store_true_params(self, attrs, props):
        params = {}
        for key, default in self.PARAMS_STORE_TRUE:
            params[key] = has_and_pop(attrs, key) or has_and_pop(props, key) or default

        return params

    @classmethod
    def name(cls):
        # todo
        # return f'Ipw{cls.__name__}'
        return cls.__name__

    def render(self, ctx, props, setup_returned):
        attrs = ctx.get('attrs', {})
        params = self._process_store_true_params(attrs, props)

        self.update_style(attrs)
        self.update_style(props)

        return self._render(ctx, attrs, props, params, setup_returned)

    def _render(self, ctx, attrs, props, params, setup_returned):
        pass


def has_and_pop(attrs, key):
    if key in attrs:
        attrs.pop(key)
        return True
    return False


def is_float(n):
    return isinstance(n, float)


def is_int(n):
    return isinstance(n, int)


def is_str(s):
    return isinstance(s, str)


def is_tuple(t):
    return isinstance(t, (tuple, list))



# ---------------------------------------------------------
# Copyright (c) vuepy.org. All rights reserved.
# ---------------------------------------------------------
from __future__ import annotations


class CompilerError(Exception):
    pass


class CompilerSyntaxError(CompilerError):
    def __init__(self, msg):
        super().__init__(f"Syntax Error: {msg}.")



from pathlib import Path

import anywidget
import traitlets
import ipywidgets as widgets

from vuepy.runtime.core.api_setup_helpers import defineEmits


class DialogWidget(anywidget.AnyWidget):
    _esm = Path(__file__).parent / 'index.js'
    _css = Path(__file__).parent / 'index.css'

    title = traitlets.Unicode('').tag(sync=True)
    width = traitlets.Unicode('50%').tag(sync=True)
    # Whether to display dialog
    value = traitlets.Bool(False).tag(sync=True)
    event = traitlets.Dict().tag(sync=True)
    # slot
    body = traitlets.List(trait=traitlets.Instance(widgets.DOMWidget)) \
        .tag(sync=True, **widgets.widget_serialization)
    footer = traitlets.List(trait=traitlets.Instance(widgets.DOMWidget)) \
        .tag(sync=True, **widgets.widget_serialization)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.observe(self._on_event, ['event'])
        self.emits = defineEmits(['open', 'close'])

    def on_close(self, callback, remove=False):
        self.emits.add_event_listener('close', callback, remove)

    def on_open(self, callback, remove=False):
        self.emits.add_event_listener('open', callback, remove)

    def _on_event(self, change):
        event = change.get("new", {})
        payload = event.get("payload")
        ev = event.get("event")
        self.emits(ev, payload)




# ---------------------------------------------------------
# Copyright (c) vuepy.org. All rights reserved.
# ---------------------------------------------------------
import base64
import pathlib
from typing import List

from IPython import InteractiveShell

from vuepy.utils.image_processer import convert_pil_image_to_bin

try:
    from PIL.ImageFile import ImageFile as PILImageFile
    from PIL.Image import Image as PILImage
    PIL_IMAGE_TYPES = (PILImageFile, PILImage)
except Exception as e:
    PIL_IMAGE_TYPES = tuple()

import ipywidgets as widgets
import markdown
from IPython.display import display

from ipywui.widgets.custom.clipboard import ClipboardWidget
from ipywui.widgets.custom.dialog import DialogWidget
from ipywui.widgets.custom.message import Message
from vuepy.log import getLogger
from vuepy.utils.common import has_changed

logger = getLogger()


class WidgetNotSupported:
    def __init__(self, *args, **kwargs):
        raise NotImplementedError(f"widget {self.__class__.__name__} not supported")


try:
    from ipywidgets import Password as _Password
except Exception as e:
    logger.warn("ipywidgets.Password import failed, fallback to Text.")
    _Password = widgets.Text

try:
    from ipywidgets import TimePicker as _TimePicker
except Exception as e:
    logger.warn(f"ipywidgets.TimePicker import failed, fallback to {WidgetNotSupported.__name__}.")

    class _TimePicker(WidgetNotSupported):
        pass

try:
    from ipywidgets import DatetimePicker as _DatetimePicker
except Exception as e:
    logger.warn(f"ipywidgets.DatetimePicker import failed, fallback to {WidgetNotSupported.__name__}.")

    class _DatetimePicker(WidgetNotSupported):
        pass

try:
    from ipywidgets import TagsInput as _TagsInput
except Exception as e:
    logger.warn(f"ipywidgets.TagsInput import failed, fallback to {WidgetNotSupported.__name__}.")

    class _TagsInput(WidgetNotSupported):
        pass

try:
    from ipywidgets import FloatsInput as _FloatsInput
except Exception as e:
    logger.warn(f"ipywidgets.FloatsInput import failed, fallback to {WidgetNotSupported.__name__}.")

    class _FloatsInput(WidgetNotSupported):
        pass

try:
    from ipywidgets import IntsInput as _IntsInput
except Exception as e:
    logger.warn(f"ipywidgets.IntsInput import failed, fallback to {WidgetNotSupported.__name__}.")

    class _IntsInput(WidgetNotSupported):
        pass

try:
    from ipywidgets import ColorsInput as _ColorsInput
except Exception as e:
    logger.warn(f"ipywidgets.ColorsInput import failed, fallback to {WidgetNotSupported.__name__}.")

    class _ColorsInput(WidgetNotSupported):
        pass

try:
    from ipywidgets import Stack as _Stack
except Exception as e:
    logger.warn(f"ipywidgets.Stack import failed, fallback to {WidgetNotSupported.__name__}.")

    class _Stack(WidgetNotSupported):
        pass


class WidgetCssStyle:
    WIDGET_STYLE_ATTR = 'style'
    WIDGET_LAYOUT_ATTR = 'layout'
    CSS_TO_WIDGET_STYLE_MAP = {
        "background_color": "button_color",
        "color": "text_color",
    }
    LAYOUT_ATTRS = {
        # size
        "height",
        "width",
        "max_height",
        "max_width",
        "min_height",
        "min_width",
        # display
        "visibility",
        "display",
        "overflow",
        # box model
        "border",
        "margin",
        "padding",
        # positioning
        "top",
        "left",
        "bottom",
        "right",
        # Flexbox
        "order",
        "flex_flow",
        "align_items",
        "flex",
        "align_self",
        "align_content",
        "justify_content",
    }

    @classmethod
    def convert_css_style_to_widget_style_and_layout(cls, css_style: str) -> dict:
        """
        :param css_style:
        :return: {'style': {...}, 'layout': {...}}
        """
        attrs_str = (kv.strip().split(':') for kv in css_style.rstrip('; ').split(';'))
        attrs = ((k.strip().replace('-', '_'), v.strip()) for k, v in attrs_str)
        ret = {
            cls.WIDGET_STYLE_ATTR: {},
            cls.WIDGET_LAYOUT_ATTR: {},
        }
        for attr, val in attrs:
            if attr in cls.LAYOUT_ATTRS:
                ret[cls.WIDGET_LAYOUT_ATTR][attr] = val
            else:
                widget_style = cls.CSS_TO_WIDGET_STYLE_MAP.get(attr, attr)
                ret[cls.WIDGET_STYLE_ATTR][widget_style] = val

        return ret

    @property
    def css_style(self):
        return self.style

    @css_style.setter
    def css_style(self, value):
        if isinstance(value, str):
            value = self.convert_css_style_to_widget_style_and_layout(value)

        style = value.get(self.WIDGET_STYLE_ATTR)
        if style:
            self.style = style

        layout = value.get('layout')
        if layout:
            self.layout = layout


class AppLayout(widgets.AppLayout, WidgetCssStyle):
    pass


class VBox(widgets.VBox, WidgetCssStyle):
    pass


class HBox(widgets.HBox, WidgetCssStyle):
    pass


class Accordion(widgets.Accordion, WidgetCssStyle):
    pass


class MarkdownViewerWidget(widgets.HTMLMath, WidgetCssStyle):
    code_highlight = pathlib.Path(__file__).parent / 'assets' / 'css' / 'md_code_highlight.css'
    with open(code_highlight) as f:
        css_style = ''.join(f.read())

    extra = [
        'markdown.extensions.extra',
        'markdown.extensions.codehilite',
        'markdown.extensions.tables',
        # 'markdown.extensions.nl2br',
    ]

    def __init__(self, value='', **kwargs):
        super().__init__(self.render(value), **kwargs)

    def render(self, md):
        html = markdown.markdown(md, extensions=self.extra)
        return f"<style>{self.css_style}</style>" + html

    def __setattr__(self, key, value):
        if key == 'value':
            value = self.render(value)

        super().__setattr__(key, value)


class BoundedFloatText(widgets.BoundedFloatText, WidgetCssStyle):
    pass


class BoundedIntText(widgets.BoundedIntText, WidgetCssStyle):
    pass


class Button(widgets.Button, WidgetCssStyle):
    CSS_TO_WIDGET_STYLE_MAP = {
        "background_color": "button_color",
        "color": "text_color",
    }
    DEFAULT_LOADING_ICON = 'fa-spinner'

    def __init__(self, **kwargs):
        self._icon_loading = f"{kwargs.get('loading-icon', self.DEFAULT_LOADING_ICON)} fa-spin"
        self._icon_org = kwargs.get('icon', '')
        loading = kwargs.pop('loading', False)
        kwargs['icon'] = self._icon_loading if loading else self._icon_org
        kwargs['description'] = kwargs.pop("label", kwargs.pop('description', ''))
        kwargs['button_style'] = kwargs.pop("type", kwargs.pop('button_style', ''))
        super().__init__(**kwargs)
        self._loading = loading

    @property
    def loading(self):
        return self._loading

    @loading.setter
    def loading(self, loading):
        if self._loading == loading:
            return
        else:
            self._loading = loading

        if loading:
            self.icon = self._icon_loading
        else:
            self.icon = self._icon_org

    @property
    def label(self):
        return self.description

    @label.setter
    def label(self, val):
        self.description = val

    @property
    def type(self):
        return self.button_style

    @type.setter
    def type(self, val):
        self.button_style = val


class Checkbox(widgets.Checkbox, WidgetCssStyle):
    CSS_TO_WIDGET_STYLE_MAP = {}

    def __init__(self, **kwargs):
        kwargs['description'] = kwargs.pop("label", kwargs.pop('description', ''))
        super().__init__(**kwargs)

    @property
    def label(self):
        return self.description

    @label.setter
    def label(self, val):
        self.description = val


class ColorPicker(widgets.ColorPicker, WidgetCssStyle):
    def __init__(self, **kwargs):
        kwargs['description'] = kwargs.pop("label", kwargs.pop('description', ''))
        super().__init__(**kwargs)

    @property
    def label(self):
        return self.description

    @label.setter
    def label(self, val):
        self.description = val


class ColorsInput(_ColorsInput, WidgetCssStyle):
    def __init__(self, **kwargs):
        kwargs['allow_duplicates'] = not kwargs.pop("unique", not kwargs.pop('allow_duplicates', True))
        super().__init__(**kwargs)

    @property
    def unique(self):
        return not self.allow_duplicates

    @unique.setter
    def unique(self, val):
        self.allow_duplicates = not val


class Combobox(widgets.Combobox, WidgetCssStyle):
    def __init__(self, **kwargs):
        kwargs['description'] = kwargs.pop("label", kwargs.pop('description', ''))
        super().__init__(**kwargs)

    @property
    def label(self):
        return self.description

    @label.setter
    def label(self, val):
        self.description = val


class Controller(widgets.Controller, WidgetCssStyle):
    pass


class DatePicker(widgets.DatePicker, WidgetCssStyle):
    def __init__(self, **kwargs):
        kwargs['description'] = kwargs.pop("label", kwargs.pop('description', ''))
        super().__init__(**kwargs)

    @property
    def label(self):
        return self.description

    @label.setter
    def label(self, val):
        self.description = val


class DateTimePicker(_DatetimePicker, WidgetCssStyle):
    def __init__(self, **kwargs):
        kwargs['description'] = kwargs.pop("label", kwargs.pop('description', ''))
        super().__init__(**kwargs)

    @property
    def label(self):
        return self.description

    @label.setter
    def label(self, val):
        self.description = val


class Dropdown(widgets.Dropdown, WidgetCssStyle):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)


class DisplayViewer(widgets.Output, WidgetCssStyle):
    def __init__(self, obj='', **kwargs):
        self.multi_thread = kwargs.pop('multi_thread', False)
        super().__init__(**kwargs)
        self.obj = obj

    def __setattr__(self, key, value):
        if key == 'obj' and has_changed(value, getattr(self, 'obj', None)):
            self.render(value)

        super().__setattr__(key, value)

    def render(self, obj):
        if not self.multi_thread:
            self.clear_output()
            with self:
                display(obj)
        else:
            self.outputs = tuple()
            self.append_display_data(obj)

    def _fix_data(self, data, metadata):
        mime_images = ['image/png', 'image/jpeg']
        for mime_image in mime_images:
            if mime_image in data and isinstance(data[mime_image], bytes):
                data[mime_image] = base64.b64encode(data[mime_image]).decode('ascii')
                break

    def append_display_data(self, display_object):
        """Append a display object as an output.

        Parameters
        ----------
        display_object : IPython.core.display.DisplayObject
            The object to display (e.g., an instance of
            `IPython.display.Markdown` or `IPython.display.Image`).
        """
        fmt = InteractiveShell.instance().display_formatter.format
        data, metadata = fmt(display_object)

        self._fix_data(data, metadata)

        self.outputs += (
            {
                'output_type': 'display_data',
                'data': data,
                'metadata': metadata
            },
        )


class FloatsInput(_FloatsInput, WidgetCssStyle):
    def __init__(self, **kwargs):
        kwargs['tag_style'] = kwargs.pop("type", kwargs.pop('tag_style', ''))
        super().__init__(**kwargs)


class FileUpload(widgets.FileUpload, WidgetCssStyle):
    pass


class FloatText(widgets.FloatText, WidgetCssStyle):
    pass


class FloatSlider(widgets.FloatSlider, WidgetCssStyle):
    pass


class FloatRangeSlider(widgets.FloatRangeSlider, WidgetCssStyle):
    pass


class FloatProgress(widgets.FloatProgress, WidgetCssStyle):
    CSS_TO_WIDGET_STYLE_MAP = {
        'color': 'bar_color',
    }

    def __init__(self, **kwargs):
        kwargs['description'] = kwargs.pop("label", kwargs.pop('description', ''))
        kwargs['bar_style'] = kwargs.pop("type", kwargs.pop('bar_style', ''))
        super().__init__(**kwargs)

    @property
    def label(self):
        return self.description

    @label.setter
    def label(self, val):
        self.description = val

    @property
    def type(self):
        return self.bar_style

    @type.setter
    def type(self, val):
        self.bar_style = val


class GridspecLayout(widgets.GridspecLayout, WidgetCssStyle):
    pass


class Label(widgets.Label, WidgetCssStyle):
    CSS_TO_WIDGET_STYLE_MAP = {
        'color': 'text_color',
    }


class HTMLMath(widgets.HTMLMath, WidgetCssStyle):
    pass


class Image(widgets.Image, WidgetCssStyle):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @staticmethod
    def _convert_to_bin_img(data):
        if isinstance(data, PIL_IMAGE_TYPES):
            return convert_pil_image_to_bin(data)
        else:
            return data

    def __setattr__(self, key, value):
        if key == 'value':
            value = self._convert_to_bin_img(value)

        super().__setattr__(key, value)


class IntText(widgets.IntText, WidgetCssStyle):
    pass


class IntSlider(widgets.IntSlider, WidgetCssStyle):
    pass


class IntRangeSlider(widgets.IntRangeSlider, WidgetCssStyle):
    pass


class IntsInput(_IntsInput, WidgetCssStyle):
    def __init__(self, **kwargs):
        kwargs['tag_style'] = kwargs.pop("type", kwargs.pop('tag_style', ''))
        super().__init__(**kwargs)


class Password(_Password, WidgetCssStyle):
    def __init__(self, **kwargs):
        kwargs['description'] = kwargs.pop("label", kwargs.pop('description', ''))
        super().__init__(**kwargs)

    @property
    def label(self):
        return self.description

    @label.setter
    def label(self, val):
        self.description = val


class Play(widgets.Play, WidgetCssStyle):
    def __init__(self, **kwargs):
        kwargs['description'] = kwargs.pop("label", kwargs.pop('description', ''))
        super().__init__(**kwargs)

    @property
    def label(self):
        return self.description

    @label.setter
    def label(self, val):
        self.description = val


class RadioButtons(widgets.RadioButtons, WidgetCssStyle):
    pass


class Select(widgets.Select, WidgetCssStyle):
    pass


class SelectMultiple(widgets.SelectMultiple, WidgetCssStyle):
    pass


class SelectionRangeSlider(widgets.SelectionRangeSlider, WidgetCssStyle):
    pass


class SelectionSlider(widgets.SelectionSlider, WidgetCssStyle):
    pass


class Stack(_Stack, WidgetCssStyle):
    def __init__(self, **kwargs):
        self.labels: List[str] = kwargs.pop('labels', [])
        selected_label = kwargs.pop('label', self.labels[0])
        kwargs['selected_index'] = self.calc_selected_index_by_label(selected_label)
        super().__init__(**kwargs)

    @property
    def selected_label(self):
        return self.labels[self.selected_index or 0]

    @selected_label.setter
    def selected_label(self, label):
        try:
            self.selected_index = self.calc_selected_index_by_label(label)
        except Exception as e:
            logger.error(f"Set Stack selected_label({label}) failed, {e}")

    def calc_selected_index_by_label(self, label):
        return self.labels.index(label)


class Tab(widgets.Tab, WidgetCssStyle):
    pass


class TagsInput(_TagsInput, WidgetCssStyle):
    def __init__(self, **kwargs):
        kwargs['allow_duplicates'] = not kwargs.pop("unique", not kwargs.pop('allow_duplicates', True))
        super().__init__(**kwargs)

    @property
    def unique(self):
        return not self.allow_duplicates

    @unique.setter
    def unique(self, val):
        self.allow_duplicates = not val


class Text(widgets.Text, WidgetCssStyle):
    def __init__(self, **kwargs):
        kwargs['description'] = kwargs.pop("label", kwargs.pop('description', ''))
        super().__init__(**kwargs)

    @property
    def label(self):
        return self.description

    @label.setter
    def label(self, val):
        self.description = val


class Textarea(widgets.Textarea, WidgetCssStyle):
    def __init__(self, **kwargs):
        kwargs['description'] = kwargs.pop("label", kwargs.pop('description', ''))
        super().__init__(**kwargs)

    @property
    def label(self):
        return self.description

    @label.setter
    def label(self, val):
        self.description = val


class TimePicker(_TimePicker, WidgetCssStyle):
    def __init__(self, **kwargs):
        kwargs['description'] = kwargs.pop("label", kwargs.pop('description', ''))
        super().__init__(**kwargs)

    @property
    def label(self):
        return self.description

    @label.setter
    def label(self, val):
        self.description = val


class ToggleButton(widgets.ToggleButton, WidgetCssStyle):
    def __init__(self, **kwargs):
        kwargs['description'] = kwargs.pop("label", kwargs.pop('description', ''))
        kwargs['button_style'] = kwargs.pop("type", kwargs.pop('button_style', ''))
        super().__init__(**kwargs)

    @property
    def label(self):
        return self.description

    @label.setter
    def label(self, val):
        self.description = val

    @property
    def type(self):
        return self.button_style

    @type.setter
    def type(self, val):
        self.button_style = val


class ToggleButtons(widgets.ToggleButtons, WidgetCssStyle):
    def __init__(self, **kwargs):
        kwargs['button_style'] = kwargs.pop("type", kwargs.pop('button_style', ''))
        super().__init__(**kwargs)

    @property
    def type(self):
        return self.button_style

    @type.setter
    def type(self, val):
        self.button_style = val


class Valid(widgets.Valid, WidgetCssStyle):
    def __init__(self, **kwargs):
        kwargs['description'] = kwargs.pop("label", kwargs.pop('description', ''))
        super().__init__(**kwargs)

    @property
    def label(self):
        return self.description

    @label.setter
    def label(self, val):
        self.description = val



# ---------------------------------------------------------
# Copyright (c) vuepy.org. All rights reserved.
# ---------------------------------------------------------
from __future__ import annotations

import abc
import dataclasses
import pathlib
from typing import Any
from typing import Callable
from typing import List
from typing import Tuple
from typing import Type

import ipywidgets as widgets
from ipywidgets import CallbackDispatcher

from vuepy import log
from vuepy.compiler_core.ast import NodeAst
from vuepy.compiler_core.ast import VueCompAst
from vuepy.compiler_core.utils import VueCompNamespace
from vuepy.compiler_dom.codegen import VueHtmlCompCodeGen
from vuepy.reactivity.effect_scope import EffectScope
from vuepy.reactivity.reactive import to_raw
from vuepy.reactivity.ref import RefImpl
from vuepy.reactivity.watch import WatchOptions
from vuepy.reactivity.watch import watch
from vuepy.runtime.core.api_lifecycle import OnBeforeMount
from vuepy.runtime.core.api_lifecycle import OnMounted
from vuepy.runtime.core.api_setup_helpers import DefineProp
from vuepy.runtime.core.api_setup_helpers import DefineProps
from vuepy.runtime.core.api_setup_helpers import defineEmits
from vuepy.runtime.core.api_setup_helpers import defineModel
from vuepy.utils.common import Nil
from vuepy.utils.common import has_changed

logger = log.getLogger()


class Dom(widgets.VBox):
    """
    https://developer.mozilla.org/en-US/docs/Web/API/Document_Object_Model/Introduction
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def appendChild(self, el):
        self.children = self.children + (el,)
        return el

    def appendLeftChild(self, el):
        self.children = (el,) + self.children
        return el


class SetupContext:
    def __init__(self):
        self.attrs = {}
        self.slots = {}
        # 触发事件
        self.emit = {}
        # 暴露公共属性
        self.expose = {}


class VNode:
    def __init__(self, type_, props, children=None, key=''):
        self.type = type_
        self.props = props
        self.children = children
        self.key = key


class VueComponent(metaclass=abc.ABCMeta):
    _name = ''
    props = {}
    template = ''

    v_model_default = 'value'

    def __init__(self, setup_ret: dict = None, template: str = '', app: "App" = None):
        self.template = template
        self._data: dict = setup_ret
        self.app = app

    @classmethod
    def name(cls):
        # return (cls._name or cls.__name__).lower()
        return cls._name or cls.__name__

    def setup(self, props: dict = None, ctx=None, vm: 'App' = None):
        """

        :param props:
        :param ctx: attrs, slots, emit
        :param vm:
        :return:
        """
        pass

    def component(self, name: str):
        comp = self.to_ns().get(name)
        try:
            if comp and (isinstance(comp, SFCFactory) or issubclass(comp, VueComponent)):
                return comp
        except Exception as e:
            logger.debug(f"find component({name}) in Component `{self.name()}` failed, {e}.")
            pass

        comp = self.app.component(name)
        return comp

    def to_ns(self):
        return {
            **self.app.config.globalProperties.to_ns(),
            **self._data,
        }

    def render(self, ctx, props, setup_returned) -> VNode:
        """
        h(tag, props|attrs, children)

        :param ctx: attrs, slots, emit
        :param setup_returned:
        :param props:
        :return:
        """
        # if not self.template:
        #     return None
        # return vue_compiler_dom(self.template)
        pass


class SFCWidgetContainer(Dom):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.on_mounted_dispatcher = CallbackDispatcher()
        self.on_before_mount_dispatcher = CallbackDispatcher()
        self.on_unmounted_dispatcher = CallbackDispatcher()
        self.on_before_unmount_dispatcher = CallbackDispatcher()


class SFC(VueComponent):
    ADD_EVENT_LISTENER_FN = '_s1_on'  # $on
    EMIT_FN = '_s1_emit'  # $emit

    def __init__(
            self,
            context: SetupContext | dict,
            props: dict,
            setup_ret: dict,
            template: str,
            app: "App",
            render: Callable[[SetupContext | dict, dict, dict], VNode] = None,
            file: str = None,
    ):
        super().__init__()
        self.app = app
        self.template = template
        self.setup_returned = setup_ret
        self._context = context
        self._props = props
        self._data: dict = setup_ret
        self._define_props = {}
        self._define_emits = {}
        self._render = render
        self.file = file
        self.scope: EffectScope = EffectScope(True)

        # setup define_x
        _emits, _on_fn = self._gen_widget_on_fn()
        _props_property = self._gen_widget_props_property()
        _model_update_events = self._gen_model_update_events()
        for _event in _model_update_events:
            _emits.add_event(_event)

        _sfc_container = type('sfc_container', (SFCWidgetContainer,), {
            **_props_property,
            self.ADD_EVENT_LISTENER_FN: _on_fn,
            self.EMIT_FN: _emits,
        })
        self.root: SFCWidgetContainer = _sfc_container()
        self._init_static_props(context.get('attrs', {}))
        self._register_lifecycle_cb(self.root)

    def _init_static_props(self, attrs):
        if not self.define_props:
            return

        _, _props = self.define_props
        for prop_name in _props.prop_names:
            val = attrs.get(prop_name, Nil)
            if val is Nil:
                continue
            setattr(self.root, prop_name, val)

    def _register_lifecycle_cb(self, container: SFCWidgetContainer):
        for _, cb in self._get_var_by_type(OnBeforeMount):
            container.on_before_mount_dispatcher.register_callback(cb.callback)

        for _, cb in self._get_var_by_type(OnMounted):
            container.on_mounted_dispatcher.register_callback(cb.callback)

    def setup(self, props: dict = None, ctx=None, vm: 'App' = None):
        return self._data

    def _gen_model_update_events(self):
        return [_model.update_event for (_, _model) in self.define_models]

    def _gen_widget_props_property(self):
        def _get(prop: DefineProp):
            def _(this):
                return prop.value

            return _

        def _set(prop: DefineProp):
            def _(this, val):
                prop.value = val

            return _

        props = [model.prop for _, model in self.define_models]
        if self.define_props:
            props.extend(self.define_props[1].props)

        return {
            prop.name: property(_get(prop), _set(prop))
            for prop in props
        }

    def _gen_widget_on_fn(self):
        emits = self.define_emits[1] if self.define_emits else defineEmits([])

        def _on_fn(this, event, callback, remove=False):
            emits.add_event_listener(event, callback, remove)

        return emits, _on_fn

    @property
    def define_props(self) -> (str, DefineProps):
        ret = self._get_var_by_type(DefineProps, limit=1)
        return ret[0] if ret else []

    @property
    def define_emits(self) -> (str, defineEmits):
        ret = self._get_var_by_type(defineEmits, limit=1)
        return ret[0] if ret else []

    @property
    def define_models(self) -> List[(str, defineModel)]:
        return self._get_var_by_type(defineModel)

    def _get_var_by_type(self, define_type, limit: int = float('inf')) -> List[Tuple[str, Any]]:
        ret = []
        count = 0
        for var_name, var in self.setup_returned.items():
            if count >= limit:
                break
            if isinstance(var, define_type):
                ret.append((var_name, var))
                count += 1
        return ret

    def __repr__(self):
        return f"{self.__class__.__name__}<{pathlib.Path(self.file).name}> at {id(self)}"

    def render(self, ctx: SetupContext = None, props: dict = None, setup_returned: dict = None) -> VNode:
        logger.info(f"🔥 Rerender {self}")
        self.scope.clear()

        # beforeMount
        self.root.on_before_mount_dispatcher()

        # compile and render
        with self.scope:
            if self._render:
                dom = self._render(self._context, self._props, self.setup_returned)
            else:
                from vuepy.compiler_sfc.compile_template import DomCompiler
                dom = DomCompiler(self, self.app).compile(self.template)

        self.root.children = [dom]

        # mounted
        self.root.on_mounted_dispatcher()

        return self.root


@dataclasses.dataclass
class SFCFactory:
    # (props: dict, context: SetupContext, app: App) -> dict | Callable[[], h]:
    setup: Callable[[dict, SetupContext | dict, 'App'], dict | Callable] = None
    template: str = ''
    # (self, ctx, props, setup_returned) -> VNode:
    render: Callable[[SetupContext | dict, dict, dict], VNode] = None
    _file: str = ''

    def gen(self, props: dict, context: SetupContext | dict, app: "App") -> "SFC":
        setup_ret = self.setup(props, context, app) if self.setup else {}
        return SFC(context, props, setup_ret, self.template, app, self.render, self._file)


class VueCompCodeGen:
    """
    h()
    """

    @classmethod
    def gen(
            cls,
            node: NodeAst,
            vm: 'VueComponent',
            ns: VueCompNamespace,
            app: "App"
    ):
        comp_ast = VueCompAst.transform(node.tag, node.attrs)
        # v-if
        if comp_ast.v_if:
            dummy = widgets.VBox()

            def _if_cond():
                return comp_ast.v_if.eval(ns)

            @watch(_if_cond, WatchOptions(immediate=True))
            def _change_v_if_widget(cond, old, on_cleanup):
                dummy.children = (cls._gen(comp_ast, node.children, vm, ns, app),) if cond else ()

            return dummy
        # v-show
        elif comp_ast.v_show:
            dummy = widgets.VBox()
            w = cls._gen(comp_ast, node.children, vm, ns, app)

            def _if_show():
                return comp_ast.v_show.eval(ns)

            @watch(_if_show, WatchOptions(immediate=True))
            def _show_widget(curr_show, old, on_cleanup):
                dummy.children = (w,) if curr_show else ()

            return dummy
        else:
            return cls._gen(comp_ast, node.children, vm, ns, app)

    @classmethod
    def _gen(
            cls,
            comp_ast: VueCompAst,
            children,
            vm: 'VueComponent',
            ns: VueCompNamespace,
            app: "App"
    ):
        component_cls: Type[VueComponent] = vm.component(comp_ast.tag)
        slots = {'default': []}
        for child in children or []:
            slot_name = getattr(child, 'v_slot', 'default')
            child = VueHtmlCompCodeGen.gen_from_fn(child) if callable(child) else child
            if slot_name == 'default':
                slots[slot_name].append(child)
            else:
                slots[slot_name] = child

        ctx = {
            'attrs': {**comp_ast.kwargs},
            'slots': slots,
            'emit': '',
        }
        props = {
            widget_attr: exp_ast.eval(ns)
            for widget_attr, exp_ast in comp_ast.v_binds.items()
        }

        if isinstance(component_cls, SFCFactory):
            component = component_cls.gen(props, ctx, app)
        else:
            component = component_cls()
        widget = component.render(ctx, props, {})

        if comp_ast.tag == 'slot' and isinstance(vm, SFC):
            _slot_name = comp_ast.kwargs.get('name', 'default')
            _slot = vm._context.get('slots', {}).get(_slot_name)
            if _slot:
                widget.children = _slot

        # v-slot
        if comp_ast.v_slot:
            widget.v_slot = comp_ast.v_slot

        # v-bind:
        for widget_attr, exp_ast in comp_ast.v_binds.items():
            if widget_attr == 'style' and hasattr(widget, 'css_style'):
                widget_attr = 'css_style'

            def _get_v_bind_value(_exp_ast=exp_ast):
                return _exp_ast.eval(ns)

            @watch(_get_v_bind_value, WatchOptions(immediate=True))
            def _v_bind_update_vm_to_view(curr, old, on_cleanup, _widget_attr=widget_attr):
                _old_val = to_raw(getattr(widget, _widget_attr, Nil))
                curr = to_raw(curr)
                if has_changed(curr, _old_val):
                    setattr(widget, _widget_attr, curr)

        # v-model:define-model=bind
        for _model_key, attr_chain in comp_ast.v_model:
            # :bind, parent to child
            # widget_attr = component_cls.v_model_default  # VueCompTag.v_model(comp_ast.tag)
            widget_attr = _model_key
            if _model_key == defineModel.DEFAULT_KEY and hasattr(component_cls, 'v_model_default'):
                widget_attr = getattr(component_cls, 'v_model_default')

            def _get_v_model_value(_attr_chain=attr_chain):
                return to_raw(ns.getattr(_attr_chain))

            @watch(_get_v_model_value, WatchOptions(immediate=True))
            def _v_model_update_vm_to_view(curr, old, on_cleanup, _widget_attr=widget_attr):
                _old_val = to_raw(getattr(widget, _widget_attr, Nil))
                curr = to_raw(curr)
                if has_changed(curr, _old_val):
                    setattr(widget, _widget_attr, curr)

            # v-on, child to parent
            def listener(_vm, _obj, _attr):
                def _(change):
                    val = change['new'] if isinstance(change, dict) else change
                    setattr(_obj, _attr, val)
                    if isinstance(_obj, defineModel) and isinstance(_vm, SFC):
                        logger.debug("%s emit(%s, %s)", _vm, _obj.update_event, val)
                        getattr(_vm.root, SFC.EMIT_FN)(_obj.update_event, val)

                return _

            obj, attr = ns.get_obj_and_attr(attr_chain)
            cls.add_event_listener(
                widget,
                f'update:{widget_attr}',
                listener(vm, obj, attr),
            )

        # v-on
        for ev, func_ast in comp_ast.v_on.items():
            def _event_handle(*args, _func_ast=func_ast, **kwargs):
                return _func_ast.eval(ns, {'__vp_args': args, '__vp_kwargs': kwargs})

            cls.add_event_listener(widget, ev, _event_handle)

        if comp_ast.v_ref:
            _ref = ns.getattr(comp_ast.v_ref)
            # if not isinstance(_ref, VueRef):
            if not isinstance(_ref, RefImpl):
                logger.error(f'v-ref={comp_ast.v_ref} needs to be of type Ref.')
            else:
                _ref.value = component if isinstance(component_cls, SFCFactory) else widget

        return widget

    @staticmethod
    def add_event_listener(widget, event, listener):
        if hasattr(widget, SFC.ADD_EVENT_LISTENER_FN):  # SFC
            getattr(widget, SFC.ADD_EVENT_LISTENER_FN)(event, listener)
        elif event.startswith('update:'):  # anywidget, ipyw
            attr = event.split(':', 1)[1]
            widget.observe(listener, names=attr)
        else:  # ipyw click
            getattr(widget, f"on_{event}")(listener)



# ---------------------------------------------------------
# Copyright (c) vuepy.org. All rights reserved.
# ---------------------------------------------------------
from __future__ import annotations

from typing import Any
from typing import Callable

from vuepy.reactivity import config
from vuepy.reactivity.dep import Dep
from vuepy.reactivity.effect import DebuggerOptions
from vuepy.reactivity.effect import ReactiveEffect
from vuepy.reactivity.ref import RefImpl
from vuepy.reactivity.ref import trackRefValue
from vuepy.reactivity.ref import triggerRefValue


ComputedGetter = Callable[[], Any]


class ComputedRefImpl(RefImpl):
    def __init__(self, getter: ComputedGetter, setter=None):
        self._dirty = True
        self.dep: Dep = None

        self.effect = ReactiveEffect(getter, scheduler=self.scheduler)
        self.effect.computed = self
        self._setter = setter
        self.is_readonly = setter is None
        self._value = None
        self.debug_msg = 'computed'

    def scheduler(self):
        if not self._dirty:
            self._dirty = True
            triggerRefValue(self)

    @property
    def value(self):
        trackRefValue(self)
        if self._dirty:
            self._dirty = False
            self._value = self.effect.run()

        return self._value

    @value.setter
    def value(self, new_val):
        if self.is_readonly:
            # todo
            print("warning")
        else:
            self._setter(new_val)


class WritableComputedOptions:
    def get(self) -> Any:
        pass

    def set(self, val: Any) -> None:
        pass


def computed(getter_or_options_or_debugger_options=None, debugger_options=None):
    is_decorator_no_arg = (getter_or_options_or_debugger_options, debugger_options) == (None, None)
    is_decorator_with_arg = isinstance(getter_or_options_or_debugger_options, DebuggerOptions)
    # @computed() or @computed(debuggerOptions)
    if is_decorator_with_arg or is_decorator_no_arg:
        def wrap(fn):
            return computed_impl(fn, getter_or_options_or_debugger_options)

        return wrap
    # @computed or computed()
    else:
        return computed_impl(getter_or_options_or_debugger_options, debugger_options)


def computed_impl(getter_or_options, debugger_options: DebuggerOptions = None) -> ComputedRefImpl:
    if isinstance(getter_or_options, WritableComputedOptions):
        getter = getter_or_options.get
        setter = getter_or_options.set
    else:
        getter = getter_or_options
        setter = None

    c_ref = ComputedRefImpl(getter, setter)
    if config.__DEV__ and debugger_options:
        c_ref.effect.onTrack = debugger_options.onTrack
        c_ref.effect.onTrigger = debugger_options.onTrigger

    return c_ref



from vuepy.utils.factory import FactoryMeta


class VuepyAppStore(metaclass=FactoryMeta):
    @classmethod
    def register(cls, app_name, app=None):
        FactoryMeta.register(cls, app_name)(app)
        return app

    @classmethod
    def get(cls, app_name):
        return cls.get_all_registry().get(app_name)



# ---------------------------------------------------------
# Copyright (c) vuepy.org. All rights reserved.
# ---------------------------------------------------------
from __future__ import annotations

import enum
from weakref import WeakKeyDictionary

from vuepy.reactivity.constant import IterateKey
from vuepy.reactivity.effect import TrackOpTypes
from vuepy.reactivity.effect import TriggerOpTypes
from vuepy.reactivity.effect import _can_reactive
from vuepy.reactivity.effect import targetMap
from vuepy.reactivity.effect import track
from vuepy.reactivity.effect import trigger
from vuepy.utils.common import Nil
from vuepy.utils.common import gen_hash_key
from vuepy.utils.common import has_changed


class ReactiveFlags(enum.Enum):
    SKIP = '_v_skip_'
    IS_REACTIVE = '_v_isReactive_'
    IS_READONLY = '_v_isReadonly_'
    IS_SHALLOW = '_v_isShallow_'
    RAW = '_v_raw_'


class ProxyStore:
    """
    proxy for target
    """

    def __init__(self):
        # { id(target): Proxy }
        self._store: dict[int, "ReactiveProxy"] = {}  # weakref.WeakValueDictionary()

    def get(self, target):
        return self._store.get(gen_hash_key(target))

    def set(self, target, proxy):
        self._store[gen_hash_key(target)] = proxy

    def delete(self, target):
        self._store.pop(gen_hash_key(target), None)

    def clear(self):
        self._store.clear()

    def len(self):
        return len(self._store)

    def __contains__(self, target):
        return gen_hash_key(target) in self._store


reactiveMap = ProxyStore()
shallowReactiveMap = WeakKeyDictionary()
readonlyMap = WeakKeyDictionary()
shallowReadonlyMap = WeakKeyDictionary()


def reactive(target) -> "ReactiveProxy" | "DictProxy" | "ListProxy":
    if isReadonly(target):
        return target

    return createReactiveObject(
        target,
        False,
        reactiveMap,
    )


class TargetType(enum.Enum):
    INVALID = 0
    COMMON = 1
    LIST = 2
    DICT = 3
    COLLECTION = 4


def getTargetType(value):
    # todo add class to common
    if isinstance(value, (list,)):
        return TargetType.LIST
    elif isinstance(value, (dict,)):
        return TargetType.DICT
    else:
        return TargetType.INVALID


class ReactiveProxy:
    ATTR_TARGET = '_vp_target_'
    ATTR_TRACK_TARGET = '_vp_track_target_'
    ATTR_SHALLOW = '_vp_shallow_'

    PROXY_ATTRS = (
        ATTR_TARGET,
        ATTR_TRACK_TARGET,
        ATTR_SHALLOW,
    )
    DICT_ATTRS = ('__class__',)

    def __init__(self, target, track_target, shallow: bool):
        self._vp_target_ = target
        self._vp_shallow_ = track_target
        self._vp_track_target_ = shallow


class DictProxy(ReactiveProxy):
    def __init__(self, target: dict, shallow: bool = False):
        super().__setattr__(ReactiveProxy.ATTR_TARGET, target)
        super().__setattr__(ReactiveProxy.ATTR_TRACK_TARGET, target)
        super().__setattr__(ReactiveProxy.ATTR_SHALLOW, shallow)

    def __delete__(self, instance):
        track_target = self._vp_track_target_
        print(f"delete {track_target}")
        targetMap.delete(track_target)
        reactiveMap.delete(track_target)

    def __getattribute__(self, name):
        if name in ReactiveProxy.PROXY_ATTRS:
            return super().__getattribute__(name)

        target = super().__getattribute__(ReactiveProxy.ATTR_TARGET)
        if name in ReactiveProxy.DICT_ATTRS:
            return getattr(target, name)

        if name not in target:
            return super().__getattribute__(name)

        track(self._vp_track_target_, TrackOpTypes.GET, name, "__getattribute__")
        res = target[name]
        return toReactive(res)

    def __setattr__(self, name, value):
        if name == ReactiveProxy.ATTR_TARGET:
            raise KeyError(f"can't set {name}")

        target = self._vp_target_
        old_value = target.get(name, Nil)
        if not self._vp_shallow_:
            if not isShallow(value):
                old_value = to_raw(old_value)
                value = to_raw(value)

            # # reactive unpack value
            # if is_ref(old_value) and not is_ref(value):
            #     old_value.value = value
            #     return

        target[name] = value
        if old_value is Nil:
            trigger(self._vp_track_target_, TriggerOpTypes.ADD, name, value, Nil, msg="__setattr__ by add")
        elif has_changed(value, old_value):
            trigger(self._vp_track_target_, TriggerOpTypes.SET, name, value, old_value, msg="__setattr__ by set")
            targetMap.delete(old_value)
            reactiveMap.delete(old_value)

    def __delattr__(self, name):
        target = self._vp_target_
        old_value = target.get(name, Nil)
        if old_value is Nil:
            return
        del target[name]
        trigger(self._vp_track_target_, TriggerOpTypes.DELETE, name, Nil, old_value, msg="delattr")

    def __getitem__(self, key):
        return self.__getattribute__(key)

    def __setitem__(self, key, value):
        self.__setattr__(key, value)

    def __delitem__(self, key):
        self.__delattr__(key)

    def __iter__(self):
        track(self._vp_track_target_, TrackOpTypes.ITER, IterateKey.DICT, msg="__iter__")
        return iter(self._vp_target_)

    def __contains__(self, key):
        track(self._vp_track_target_, TrackOpTypes.HAS, key, msg='__contains__')
        return key in self._vp_target_

    def __len__(self):
        track(self._vp_track_target_, TrackOpTypes.ITER, IterateKey.DICT, msg="__len__")
        return len(self._vp_target_)

    def __repr__(self):
        track(self._vp_track_target_, TrackOpTypes.ITER, IterateKey.DICT, msg="__repr__")
        return repr(self._vp_target_)

    def __str__(self):
        track(self._vp_track_target_, TrackOpTypes.ITER, IterateKey.DICT, msg="__str__")
        return str(self._vp_target_)

    def __format__(self, format_spec):
        track(self._vp_track_target_, TrackOpTypes.ITER, IterateKey.DICT, msg="__format__")
        return self._vp_target_.__format__(format_spec)

    def clear(self):
        ret = self._vp_target_.clear()
        for key in list(self._vp_target_.keys()):
            trigger(self._vp_track_target_, TriggerOpTypes.CLEAR, key, Nil, msg="clear")
        return ret

    def copy(self):
        track(self._vp_track_target_, TrackOpTypes.ITER, IterateKey.DICT, msg="copy")
        return self._vp_target_.copy()

    def get(self, key, default=None):
        track(self._vp_track_target_, TrackOpTypes.GET, key, msg="get")
        res = self._vp_target_.get(key, default)
        return toReactive(res)

    def keys(self):
        track(self._vp_track_target_, TrackOpTypes.ITER, IterateKey.DICT, msg="keys")
        return self._vp_target_.keys()

    def values(self):
        track(self._vp_track_target_, TrackOpTypes.ITER, IterateKey.DICT, msg="values")
        return (toReactive(value) for value in self._vp_target_.values())

    def items(self):
        track(self._vp_track_target_, TrackOpTypes.ITER, IterateKey.DICT, msg="items")
        return ((k, toReactive(v)) for k, v in self._vp_target_.items())

    def pop(self, key, default=None):
        ret = self._vp_target_.pop(key, default)
        trigger(self._vp_track_target_, TriggerOpTypes.DELETE, key, None, msg="pop")
        return ret

    def popitem(self):
        key, value = self._vp_target_.popitem()
        trigger(self._vp_track_target_, TriggerOpTypes.DELETE, key, None, msg="popitem")
        return key, value

    def setdefault(self, key, default=None):
        if key in self._vp_target_:
            return
        ret = self.__setattr__(key, default)
        return ret

    def update(self, other):
        for key, value in other.items():
            self.__setattr__(key, value)


class ListProxy(ReactiveProxy):
    def __init__(self, target: list, shallow: bool = False):
        super().__setattr__(ReactiveProxy.ATTR_TARGET, target)
        super().__setattr__(ReactiveProxy.ATTR_TRACK_TARGET, target)
        super().__setattr__(ReactiveProxy.ATTR_SHALLOW, shallow)

    def __getitem__(self, index):
        """
        :param index: int, slice
        :return: ReactiveProxy or Any
        """
        target = self._vp_target_
        res = target[index]
        track(self._vp_track_target_, TrackOpTypes.ITER, index, "list.__getitem__")
        return toReactive(res)

    def __setitem__(self, index, value):
        target = self._vp_target_
        old_value = target[index]

        if not has_changed(value, old_value):
            return

        target[index] = value
        trigger(self._vp_track_target_, TriggerOpTypes.SET, index, value, old_value,
                msg="__setitem__ by set")
        targetMap.delete(old_value)
        reactiveMap.delete(old_value)

    def __delitem__(self, index):
        target = self._vp_target_
        old_value = target[index]
        if old_value is Nil:
            return
        del target[index]
        trigger(self._vp_track_target_, TriggerOpTypes.ITER, IterateKey.LIST, msg="__delitem__")

    def __iter__(self):
        track(self._vp_track_target_, TrackOpTypes.ITER, IterateKey.LIST, msg="__iter__")
        return iter(toReactive(item) for item in self._vp_target_)

    def __contains__(self, item):
        track(self._vp_track_target_, TrackOpTypes.ITER, item, msg='__contains__')
        return item in self._vp_target_

    def __len__(self):
        track(self._vp_track_target_, TrackOpTypes.ITER, IterateKey.LIST, msg="__len__")
        return len(self._vp_target_)

    def __repr__(self):
        track(self._vp_track_target_, TrackOpTypes.ITER, IterateKey.LIST, msg="__repr__")
        return f"{self.__class__.__name__}: {repr(self._vp_target_)}"

    def __str__(self):
        track(self._vp_track_target_, TrackOpTypes.ITER, IterateKey.LIST, msg="__str__")
        return f"{self.__class__.__name__}: {str(self._vp_target_)}"

    def append(self, item):
        ret = self._vp_target_.append(item)
        trigger(self._vp_track_target_, TriggerOpTypes.ADD, len(self._vp_target_), item,
                msg="append")
        return ret

    def clear(self):
        ret = self._vp_target_.clear()
        trigger(self._vp_track_target_, TriggerOpTypes.CLEAR, None, msg="clear")
        return ret

    def copy(self):
        track(self._vp_track_target_, TrackOpTypes.ITER, IterateKey.LIST, msg="copy")
        return self._vp_target_.copy()

    def count(self, item):
        track(self._vp_track_target_, TrackOpTypes.ITER, IterateKey.LIST, msg="count")
        return self._vp_target_.count(item)

    def extend(self, iterable):
        self._vp_target_.extend(iterable)
        trigger(self._vp_track_target_, TriggerOpTypes.ADD, IterateKey.LIST, msg="extend")

    def index(self, item, start=0, end=None):
        track(self._vp_track_target_, TrackOpTypes.ITER, item, msg="index")
        return self._vp_target_.index(item, start, end)

    def insert(self, index, item):
        ret = self._vp_target_.insert(index, item)
        trigger(self._vp_track_target_, TriggerOpTypes.ADD, IterateKey.LIST, msg="insert")
        return ret

    def pop(self, index=-1):
        ret = self._vp_target_.pop(index)
        trigger(self._vp_track_target_, TriggerOpTypes.DELETE, IterateKey.LIST, msg="pop")
        return ret

    def remove(self, item):
        index = self._vp_target_.index(item)
        return self.__delitem__(index)

    def reverse(self):
        ret = self._vp_target_.reverse()
        trigger(self._vp_track_target_, TriggerOpTypes.ITER, IterateKey.LIST, msg="reverse")
        return ret

    def sort(self, key=None, reverse=False):
        ret = self._vp_target_.sort(key=key, reverse=reverse)
        trigger(self._vp_track_target_, TriggerOpTypes.ITER, IterateKey.LIST, msg="sort")
        return ret


TARGET_PROXY_MAP = {
    TargetType.COMMON: ReactiveProxy,
    TargetType.DICT: DictProxy,
    TargetType.LIST: ListProxy,
}


def createReactiveObject(target, is_readonly: bool, proxyMap: ProxyStore) -> ReactiveProxy:
    if not _can_reactive(target):
        # todo __DEV__ log
        return target

    # target is already a Proxy, return it
    if isinstance(target, ReactiveProxy):
        return target

    # target already has corresponding Proxy
    # can't use the `if exist_proxy` method because `if` will call the __len__ method to track
    if target in proxyMap:
        return proxyMap.get(target)

    # only specific value types can be observed
    target_type = getTargetType(target)
    if target_type == TargetType.INVALID:
        return target

    _proxy = TARGET_PROXY_MAP[target_type](target)
    proxyMap.set(target, _proxy)
    return _proxy


def toReactive(value):
    return reactive(value) if _can_reactive(value) else value


def to_raw(proxy):
    if isinstance(proxy, ReactiveProxy):
        return proxy._vp_target_

    return proxy


toRaw = to_raw


def is_reactive(value) -> bool:
    return isinstance(value, ReactiveProxy)


isReactive = is_reactive


def shallow_reactive(target):
    pass


def shallow_readonly(target):
    pass


def isShallow(value) -> bool:
    return False


def isReadonly(value) -> bool:
    return False


def isProxy(value):
    return is_reactive(value) or isReadonly(value)



# ---------------------------------------------------------
# Copyright (c) vuepy.org. All rights reserved.
# ---------------------------------------------------------
from __future__ import annotations

from vuepy import log
from vuepy.reactivity import config
from vuepy.reactivity import effect
from vuepy.reactivity.effect import DepStore
from vuepy.reactivity.effect import TrackOpTypes
from vuepy.reactivity.effect import TriggerOpTypes
from vuepy.reactivity.effect import trackEffects
from vuepy.reactivity.effect import triggerEffects
from vuepy.reactivity.reactive import DictProxy
from vuepy.reactivity.reactive import ListProxy
from vuepy.reactivity.reactive import isReadonly
from vuepy.reactivity.reactive import isShallow
from vuepy.reactivity.reactive import toReactive
from vuepy.reactivity.reactive import to_raw
from vuepy.utils.common import has_changed

logger = log.getLogger()

g_DEP_STORE = DepStore()


def ref(value, debug_msg='') -> "RefImpl":
    return createRef(value, False, debug_msg=debug_msg)


def shallowRef(value) -> "ShallowRef":
    return createRef(value, True)


def createRef(raw_value, shallow: bool, debug_msg=''):
    if isRef(raw_value):
        return raw_value

    return ShallowRef(raw_value, debug_msg) if shallow else Ref(raw_value, debug_msg)


class RefImpl:
    def __init__(self, value, is_shallow, debug_msg=''):
        self._is_shallow = is_shallow
        self._raw_value = value if is_shallow else to_raw(value)
        self._value = value if is_shallow else toReactive(value)
        self.debug_msg = debug_msg

    @property
    def value(self):
        trackRefValue(self)
        return self._value

    @value.setter
    def value(self, new_val):
        use_direct_value = self._is_shallow or isShallow(new_val) or isReadonly(new_val)
        new_val = new_val if use_direct_value else to_raw(new_val)
        logger.debug('%s set value %s to %s', self, self._raw_value, new_val)
        if has_changed(new_val, self._raw_value):
            self._raw_value = new_val
            self._value = new_val if use_direct_value else toReactive(new_val)
            triggerRefValue(self, new_val)

    def __repr__(self):
        return f"RefImpl:{self.debug_msg} at {id(self)}"


class Ref(RefImpl):
    def __init__(self, value, debug_msg=''):
        super().__init__(value, False, debug_msg)


class ShallowRef(RefImpl):
    def __init__(self, value, debug_msg=''):
        super().__init__(value, True, debug_msg)


def trackRefValue(ref):
    if effect.shouldTrack and effect.activeEffect:
        dep = g_DEP_STORE.get_or_create(ref)
        debugger_event = config.__DEV__ and {
            'target': ref,
            'type': TrackOpTypes.GET,
            'key': 'value',
        }
        trackEffects(dep, debugger_event)


def triggerRefValue(ref, new_val=None):
    dep = g_DEP_STORE.get(ref)
    if not dep:
        return

    debugger_event = config.__DEV__ and {
        'target': ref,
        'type': TriggerOpTypes.SET,
        'key': 'value',
        'newValue': new_val,
    }
    triggerEffects(dep, debugger_event)


def triggerRef(ref_obj: ShallowRef):
    triggerRefValue(ref_obj, ref_obj.value if config.__DEV__ else None)


# todo add Readonly
class GetterRefImpl(Ref):
    def __init__(self, _getter):
        self._getter = _getter
        self.debug_msg = 'GetterRefImpl'

    @property
    def value(self):
        return self._getter()


class ObjectRefImpl(Ref):
    def __init__(self, _object, _key, _default_val):
        self._object = _object
        self._key = _key
        self._default_val = _default_val

    @property
    def value(self):
        return self._object.get(self._key, self._default_val)

    @value.setter
    def value(self, new_val):
        self._object[self._key] = new_val


def isRef(val) -> bool:
    return isinstance(val, RefImpl)


def unref(ref_obj):
    return ref_obj.value if isRef(ref_obj) else ref_obj


def toRef(source, key: str = '', default_val=None) -> Ref:
    if isRef(source):
        return source
    elif callable(source):
        return GetterRefImpl(source)
    elif hasattr(source, key):
        return propertyToRef(source, key, default_val)
    else:
        return ref(source)


def propertyToRef(source, key, default_val=None):
    val = source[key]
    return val if isRef(val) else ObjectRefImpl(source, key, default_val)


def toValue(source):
    return source() if callable(source) else unref(source)


def toRefs(obj):
    if isinstance(obj, ListProxy):
        ret = []
        for i in range(len(obj)):
            ret.append(propertyToRef(obj, i))
        return ret
    elif isinstance(obj, DictProxy):
        ret = {}
        for key in obj:
            ret[key] = propertyToRef(obj, key)
        return ret
    else:
        logger.warn(f"toRefs() expects a reactive object but received a plain one.")
        return {}



# ---------------------------------------------------------
# Copyright (c) vuepy.org. All rights reserved.
# ---------------------------------------------------------
from ipywui.comps import *
from ipywui.core import IPywidgetsComponent
from ipywui.core import wui
from vuepy.version import VERSION

__version__ = VERSION



# ---------------------------------------------------------
# Copyright (c) vuepy.org. All rights reserved.
# ---------------------------------------------------------
import io


def convert_pil_image_to_bin(img, format=None) -> bytes:
    """

    :param img: PIL.Image
    :param format: PNG
    :return:
    """
    with io.BytesIO() as buf:
        img.save(buf, format=format or img.format)
        return buf.getvalue()


def convert_opencv_image_to_bin(img, format='.png') -> bytes:
    """

    :param img: opencv img(ndarray)
    :param format: .png .jpg
    :return:
    """
    import cv2 as cv
    _, img_encoded = cv.imencode(format, img)
    with io.BytesIO(img_encoded.tobytes()) as buf:
        return buf.getvalue()



# coding: utf-8
import json
import logging
from pathlib import Path

import IPython
from IPython.core.magic import register_line_cell_magic
from IPython.core.magic import register_line_magic

from ipywui import wui
from vuepy import log
from vuepy.compiler_sfc.compile import SFCFile
from vuepy.runtime.core.api_create_app import create_app
from vuepy.runtime.core.import_sfc import import_sfc
from vuepy.utils.appstore import VuepyAppStore


def get_curr_ipynb_dir():
    ipy = IPython.get_ipython()
    if '__vsc_ipynb_file__' in ipy.user_ns:
        return Path(ipy.user_ns['__vsc_ipynb_file__']).parent
    else:
        return Path().absolute()


class LogLevelScope:
    def __init__(self, logger: logging.Logger, level=logging.ERROR):
        self.logger = logger
        self.org_level = logger.level
        self.level = level

    def __enter__(self):
        self.logger.setLevel(self.level)

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.logger.setLevel(self.org_level)


@register_line_cell_magic
def vuepy_import(vue_sfc, cell=''):
    """
    import Component from sfc_file or raw_content of sfc

    usage:

    ```python
    from vuepy import create_app
    from vuepy.utils import magic

    App=%vuepy_import test.vue
    app = create_app(App)
    app.mount()
    ```

    usage:
    ```
    %%vuepy_import Component1
    <template>
      <Button description="add"
              button_style="info"
      ></Button>
    </template>
    ```

    :param vue_sfc:
    :param cell:
    :return: App
    """
    if cell:
        component_var_name = vue_sfc.strip()
        ipython = IPython.get_ipython()
        ipython.user_ns[component_var_name] = import_sfc(cell, raw_content=True)
        print(f"import Component {component_var_name} success.")
    else:
        return import_sfc(vue_sfc)


@register_line_magic
def vuepy_demo(vue_sfc):
    """
    %vuepy_demo test.vue

    :param vue_sfc:
    :return:
    """
    args = vue_sfc.strip().split()
    if len(args) == 2:
        vue_sfc, return_app = args
    else:
        return_app = 'no'

    sfc_file_path = Path(vue_sfc.strip())
    sfc_file = SFCFile.load(sfc_file_path)

    script_content = 0
    if sfc_file.script_src:
        _script_file = sfc_file.file.parent / sfc_file.script_src
        with open(_script_file) as f:
            _comment = f'# {sfc_file.script_src}\n\n'
            script_content = _comment + f.read()

    sfc_file_comment = f'<!-- {sfc_file_path} -->\n'
    print(json.dumps({
        'vue': sfc_file_comment + sfc_file.content,
        'setup': script_content,
    }))

    with LogLevelScope(log.getLogger()):
        App = import_sfc(sfc_file_path)
        app = create_app(App).use(wui)
        if return_app != 'no':
            return app
        else:
            return app.mount()


@register_line_magic
def vuepy_log(cmd):
    """
    show log.

    :param cmd: clear
    :return:
    """
    if cmd == 'clear':
        log.logout.clear_output()
    return log.logout


@register_line_cell_magic
def vuepy_run(vue_file, cell=''):
    """
    run Vuepy app from vue_file or raw_content of vue file or Component

    usage:
    cell magic: from raw_content of vue file
    ```
    ------------------
    %%vuepy_run
    <template>
      <Button description="add"
              button_style="info"
      ></Button>
    </template>
    ------------------
    ```

    line magic: from vue_file
    ```
    ------------------
    %vuepy_run app.vue
    ------------------
    ```

    line magic: from Component
    ```
    ------------------
    %%vuepy_import Component1
    <template>
      <Button description="add"
              button_style="info"
      ></Button>
    </template>

    ------------------
    # Equivalent to create_app(Component1).mount()
    %vuepy_run $$Component1

    ------------------
    ```

    :param vue_file: vue_file | Component
    :param cell: raw_content of vue file | None
    :return:
    """
    if cell:
        App = import_sfc(cell, raw_content=True)
    else:
        vue_file = vue_file.strip()
        if vue_file.startswith('$'):
            ipython = IPython.get_ipython()
            App = ipython.user_ns[vue_file[1:]]
        elif vue_file in VuepyAppStore.get_all_registry():
            App = VuepyAppStore.get(vue_file)
        else:
            App = import_sfc(vue_file)

    app = create_app(App)
    return app.mount()


def load_ipython_extension(ipython):
    def vuepy_run_complete(self, event):
        return VuepyAppStore.get_all_registry().keys()

    ipython.set_hook('complete_command', vuepy_run_complete, re_key='%vuepy_run')


load_ipython_extension(IPython.get_ipython())
